#include <time.h>
#include <fftw3.h> 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#define REAL 0
#define IMAG 1

void  print_time(const char * message, double startplan, double endplan) {
  double time_plan = (endplan - startplan)/(double)CLOCKS_PER_SEC;

  /* output of the computation time */
  double minutes = floor(time_plan / 60);
  double seconds = floor(time_plan - minutes*60);
  double millisec = (time_plan - minutes*60 - seconds) * 1000;

  printf("%s\n", message);
  printf("%f", minutes);
  printf("%s", " min. ");
  printf("%f", seconds);
  printf("%s", " sec. ");
  printf("%f", millisec);
  printf("%s\n\n", " millisec. ");

}

/* 
 * Associa ad un file su disco (tipicamente double, long double, complex, ecc.)
 * un array in memoria. Il file deve contenere in formato binario il contenuto
 * desiderato dell'array e deve essere di dimensione multipla della dimensione
 * del tipo desiderato.
 * Il file (e l'array) e` in sola lettura;
 * dopo l'uso va' disallocata la memoria (vedi unmapRO_File2Array())
 * Ritorna un puntatore all'array e come argomento il file descriptor 
 * del file aperto (da chiudere successivamente)
 * La dimensione dell'array che ci si aspetta e` fornita come argomento,
 * se non combacia la funzione fallisce.
 */
void * mapRO_File2Array(const char *filename, int* fd, size_t arraySize, size_t typeSize) {
  char* funcName = "mapRO_File2Array()";
  
  *fd = -1;
  struct stat st;
  void* mmappedData;
  size_t fileSize;

  if (stat(filename, &st) != 0) {
    fprintf(stderr, "%s: stat(%s) failed: %s\n", funcName, filename, strerror(errno));
    return NULL;
  }
  
  if (st.st_size % typeSize != 0) {
    fprintf(stderr, "%s: size of %s doesn't match 'typeSize' (%ld) multiple size\n", funcName, filename, typeSize);
    return NULL;
  }

  if (arraySize > st.st_size / typeSize) {
    fprintf(stderr, "%s: size of %s lower than desidered array size (%ld < %ld/%ld)\n", funcName, filename, arraySize, st.st_size, typeSize);
    return NULL;
  }

  fileSize = arraySize * typeSize;
  
  if ((*fd = open(filename, O_RDONLY, 0)) == -1) {
    fprintf(stderr, "%s: open(%s) failed: %s\n", funcName, filename, strerror(errno));
    return NULL;
  }
  
  if ((mmappedData = mmap(NULL, fileSize, PROT_READ, MAP_PRIVATE | MAP_POPULATE, *fd, 0)) == MAP_FAILED) {
    fprintf(stderr, "%s: mmap(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }

  return mmappedData;
}

/*
 * opens for read/write an non-existing file (or empty an existing one)
 */
void * mapRWCreate_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize) {
  char * funcName = "mapRWCreate_File2Array()";

  int rlseek = -1;
  int rwrite = -1;
  char* zero = NULL;
  void* mmappedData = NULL;
  size_t fileSize;

  *fd = -1;
  if ((zero = calloc(1, typeSize)) == NULL) {
    fprintf(stderr, "%s: calloc(%ld) failed.\n", funcName, typeSize);
    return NULL;
  }

  if ((*fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, (mode_t)0600)) == -1) {
    fprintf(stderr, "%s: open(%s) failed: %s\n", funcName, filename, strerror(errno));
    return NULL;
  }

  fileSize = arraySize * typeSize;
  
  if ((rlseek = lseek(*fd, (arraySize - 1) * typeSize, SEEK_SET)) == -1) {
    fprintf(stderr, "%s: lseek(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }

  if ((rwrite = write(*fd, &zero, typeSize)) == -1) {
    fprintf(stderr, "%s: write(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }
  
  free(zero);

  /* if ((mmappedData = mmap(NULL, fileSize, PROT_READ | PROT_WRITE, MAP_SHARED, *fd, 0)) == MAP_FAILED) { */
  if ((mmappedData = mmap(NULL, fileSize, PROT_READ | PROT_WRITE, MAP_SHARED | MAP_NORESERVE , *fd, 0)) == MAP_FAILED) {
    fprintf(stderr, "%s: mmap(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }

  return mmappedData;
}

/* 
 * open for read/write an existing file 
 */
void * mapRW_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize) {
  char* funcName = "mapRW_File2Array()";
    
  int rlseek = -1;
  int rwrite = -1;
  char* zero = NULL;
  void* mmappedData = NULL;
  size_t fileSize;

  *fd = -1;
  if ((zero = calloc(1, typeSize)) == NULL) {
    fprintf(stderr, "%s: calloc(%ld) failed.\n", funcName, typeSize);
    return NULL;
  }

  if ((*fd = open(filename, O_RDWR, (mode_t)0600)) == -1) {
    fprintf(stderr, "%s: open(%s) failed: %s\n", funcName, filename, strerror(errno));
    return NULL;
  }

  fileSize = arraySize * typeSize;
  
  if ((rlseek = lseek(*fd, (arraySize - 1) * typeSize, SEEK_SET)) == -1) {
    fprintf(stderr, "%s: lseek(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }

  if ((rwrite = write(*fd, &zero, typeSize)) == -1) {
    fprintf(stderr, "%s: write(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }
  
  if ((mmappedData = mmap(NULL, fileSize, PROT_READ | PROT_WRITE, MAP_SHARED, *fd, 0)) == MAP_FAILED) {
    fprintf(stderr, "%s: mmap(%s) failed: %s\n", funcName, filename, strerror(errno));
    close(*fd);
    return NULL;
  }

  return mmappedData;
}

void unmap_File2Array(void * mmappedData, size_t filesize, int fd) {
  const char* funcName = "unmap_File2Array()";
  int rc = munmap(mmappedData, filesize);
  if (rc != 0) {
    fprintf(stderr, "%s: munmap() failed: %s\n", funcName, strerror(errno));
    exit(1);
  }
  close(fd);
}

char * mallocFilename(const char* basename, const char* suffix) {
  char* funcName = "mallocFilename()";
  char* filename = NULL;
  size_t len = 0;
  size_t written = 0;
  
  if (basename == NULL) {
    fprintf(stderr, "%s: WARNING: basename == NULL\n", funcName);
  }

  if (suffix == NULL) {
    fprintf(stderr, "%s: WARNING: suffix == NULL\n", funcName);
  }

  if (basename == NULL && suffix == NULL) {
    fprintf(stderr, "%s: ERROR: basename == NULL && suffix == NULL\n", funcName);
    return NULL;
  }

  len = strlen(basename) + strlen(suffix);
  if (len > 255) {
    fprintf(stderr, "%s: ERROR: filename %s%s too big\n", funcName, basename, suffix);
    return NULL;
  }

  if ((filename = malloc(len+2)) == NULL) {
    fprintf(stderr, "%s: malloc(%ld) failed!\n", funcName, len+2);
    return NULL;
  }

  if ((written = snprintf(filename, len+1, "%s%s", basename, suffix)) != len) {
    fprintf(stderr, "%s: ERROR snprintf() returns %ld instead of %ld\n", funcName, written, len);
    return NULL;
  }

  return filename;
  
}

#define ARRAYSIZE 1024
void readLongDoubleValues(long double* data, unsigned long int q, int check_for_q, const char * filename) {
    FILE *myFile = NULL;
    unsigned long int i;
    unsigned long int dummy;
    long double array[ARRAYSIZE];

    long counter;
    long pointer;

    if (data == NULL) {
      fprintf(stderr, "data is null!\n");
      exit(1);
    }

    if ((myFile = fopen(filename, "r")) == NULL) {
      fprintf(stderr, "Can't open %s: %s\n", filename, strerror(errno));
      exit(1);
    }

    if (check_for_q == 1) {
      if (fread(array, sizeof(long double), 1, myFile) == 1) {
	dummy = (unsigned long) array[0];
	if (dummy != q) {
	  fprintf(stderr, "q=%lu, dummy=%lu. Giving up!\n", q, dummy);
	  exit(1);
	}
      }
    }
    
    pointer = 0;
    counter = 0;
    while ((counter = fread(array, sizeof(long double), ARRAYSIZE, myFile)) > 0) {
      for (i = 0; i < counter; i++) {
	data[pointer] = (double) array[i];
	pointer++;
      }
    }
    fclose(myFile);
    
    printf("\nRead %ld doubles\n", pointer);    
}

void writeLongDoubleValues(long double *data, unsigned long int size, const char *filename) {
  FILE *myFile = NULL;
  unsigned long int writed = 0;

  if (data == NULL) {
    fprintf(stderr, "writeLongDoubleValues(): data == NULL. Giving up!\n");
    exit(1);
  }

  if ((myFile = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "writeLongDoubleValues(): Can't open file %s: %s\n", filename, strerror(errno));
    exit(1);
  }

  if ((writed = fwrite(data, sizeof(long double), size, myFile)) != size) {
    fprintf(stderr, "writeLongDoubleValues(): Can't write array on file %s: %s\n", filename, strerror(errno));
    exit(1);
  }

  fclose(myFile);
  printf("\nWrited %ld doubles\n", writed);    
}

void read_FFTWL_Complex_Array(fftwl_complex *data, unsigned long int size, const char *filename) {
  FILE *myFile = NULL;
  unsigned long int read = 0;

  if (data == NULL) {
    fprintf(stderr, "read_FFTWL_Comples_Array(): data == NULL. Giving up!\n");
    exit(1);
  }

  if ((myFile = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "Can't open file %s: %s\n", filename, strerror(errno));
    exit(1);
  }

  if ((read = fread(data, sizeof(fftwl_complex), size, myFile)) != size) {
    fprintf(stderr, "Can't read array on file %s: %s\n", filename, strerror(errno));
    exit(1);
  }

  fclose(myFile);
}

void writearray(void * data, unsigned long int size, size_t typesize, const char *filename) {
  const char * funcname = "writearray()";
  FILE *myFile = NULL;
  unsigned long int writed = 0;

  if (data == NULL) {
    fprintf(stderr, "%s: data == NULL. Giving up!\n", funcname);
    exit(1);
  }

  if (filename == NULL) {
    fprintf(stderr, "%s: filename == NULL. Giving up!\n", funcname);
    exit(1);
  }

  if ((myFile = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "%s: Can't open file %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }

  if ((writed = fwrite(data, typesize, size, myFile)) != size) {
    fprintf(stderr, "Can't write array on file %s: %s\n", filename, strerror(errno));
    exit(1);
  }

  fclose(myFile);
  
}

/* 
   Dump a fftwl_complex array to a filename
 */
void * mkmmapcopy(void * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix, const char mode, int *fd) {
  const char * funcname = "mkmmapcopy()";

  if (a == NULL) {
    fprintf(stderr, "%s: input array == NULL. Giving up\n", funcname);
    exit(1);
  }
  
  if (n < 1) {
    fprintf(stderr, "%s: input array size < 1. Giving up\n", funcname);
    exit(1);
  }

  if (basename == NULL) {
    fprintf(stderr, "%s: input basename == NULL. Giving up\n", funcname);
    exit(1);
  }

  if (suffix == NULL) {
    fprintf(stderr, "%s: input suffix == NULL. Giving up\n", funcname);
    exit(1);
  }

  if (mode != 'r' && mode != 'w' && mode != '-') {
    fprintf(stderr, "%s: input parameter mode not 'r', not 'w' and not '-'. Giving up\n", funcname);
    exit(1);    
  }

  char * filename = NULL;
  size_t dim = strlen(basename) + strlen(suffix) + 2;
  if ((filename = malloc(dim * sizeof(char))) == NULL) {
    fprintf(stderr, "%s: unable to allocate filename. Giving up\n", funcname);
    exit(1);
  }

  snprintf(filename, dim, "%s%s", basename, suffix);
  
  clock_t startcopy = clock();
  writearray(a, n, typesize, filename);
  clock_t endcopy = clock();
  print_time(filename, startcopy, endcopy);
  fflush(stdout);
  
  if (mode == 'r') {
    return mapRO_File2Array(filename, fd, n, typesize);
  }

  if (mode == 'w') {
    return mapRW_File2Array(filename, fd, n, typesize);
  }
  
  /* do no mmap file (only copy) */
  return NULL;
}

fftwl_complex * mapfile(const char * name, const char * basename, long int size, int * fd) {
  char * funcname = "mapfile()";
  
  if (name == NULL) {
    fprintf(stderr, "%s: name == NULL. Giving up\n", funcname);
    exit(1);
  }

  if (basename == NULL) {
    fprintf(stderr, "%s: basename == NULL. Giving up\n", funcname);
    exit(1);
  }

  if (size < 1) {
    fprintf(stderr, "%s: size < 1. Giving up\n", funcname);
    exit(1);
  }

  if (fd == NULL) {
    fprintf(stderr, "%s: fd == NULL. Giving up\n", funcname);
    exit(1);
  }

  char * filename = NULL;
  size_t dim = strlen(basename) + strlen(name) + 1;
  if ((filename = malloc(dim * sizeof(char))) == NULL) {
    fprintf(stderr, "%s: unable to allocate filename. Giving up\n", funcname);
    exit(1);
  }
  snprintf(filename, dim, "%s%s", basename, name);

  return mapRO_File2Array(filename, fd, size, sizeof(fftwl_complex));  
}

/* 
   copy the data for array a in a long double file
   the copy is coniugated (x+iy -> x-iy) and divided
   by n (the size). This function in preliminary to
   the computation of iFFT (inverse of the FFT)
*/
void * coniugateAndDivideFFTFileInMappedFile(fftwl_complex * a, ptrdiff_t n, size_t typesize,
			    const char * basename, const char * suffix) {
  const char * funcname = "copyFileInMappedFile()";

  fftwl_complex * input = NULL;
  int fd;
  long double factor = n;
  
  if (a == NULL) {
    fprintf(stderr, "%s: destination array == NULL. Giving up\n", funcname);
    exit(1);
  }
  
  if (n < 1) {
    fprintf(stderr, "%s: destination array size < 1. Giving up\n", funcname);
    exit(1);
  }

  if (basename == NULL) {
    fprintf(stderr, "%s: input basename == NULL. Giving up\n", funcname);
    exit(1);
  }

  if (suffix == NULL) {
    fprintf(stderr, "%s: input suffix == NULL. Giving up\n", funcname);
    exit(1);
  }

  if ((input = mapfile(suffix, basename, n, &fd)) == NULL) {
    fprintf(stderr, "%s: input suffix == NULL. Giving up\n", funcname);
    exit(1);    
  }

  for (ptrdiff_t i = 0; i < n; i++) {
    a[i][REAL] =   input[i][REAL] / factor;
    a[i][IMAG] = - input[i][IMAG] / factor;
  }

  return a;
}

void prntc(fftwl_complex* result, unsigned long int NUM_POINTS) { /* for checking results */
    int i;
    for (i = 0; i < NUM_POINTS; ++i) {
        printf("%Lf\n", result[i][REAL]);
        printf("%Lf\n", result[i][IMAG]);
    }
}

void prntr(long double* result,unsigned long int NUM_POINTS) { /* for checking results */
    int i;
    for (i = 0; i < NUM_POINTS; ++i) {
        printf("%Lf\n", result[i]);
    }
}

/* stampa un vettore in formato matlab/octave a scopo di debug */
void printcomplexvector(fftwl_complex * v, ptrdiff_t start, ptrdiff_t step, ptrdiff_t size, const char * name, FILE *stream) {
  ptrdiff_t k = 0;

  fprintf(stream, "%s = [ ", name);
  for(k = start; k < size; k+=step) {
    fprintf(stream, "%LE %+LEi", v[k][REAL], v[k][IMAG]);
    if (k + step >= size) {
      fprintf(stream, " ]\n");
    } else {
      fprintf(stream, ", ");
    }
  }
}
